#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from neptune.new.internal.backends.utils import build_operation_url


class TestNeptuneBackendMock(unittest.TestCase):
    def test_building_operation_url(self):
        urls = {
            build_operation_url(
                'https://app.stage.neptune.ai',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'https://app.stage.neptune.ai',
                '/api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'https://app.stage.neptune.ai/',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'https://app.stage.neptune.ai/',
                '/api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai',
                '/api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai/',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai/',
                '/api/leaderboard/v1/attributes/download'
            ),
        }
        self.assertEqual({'https://app.stage.neptune.ai/api/leaderboard/v1/attributes/download'}, urls)
